#
# bibliography package for Perl
#
# text routines
#
# This reads raw text and does no processing on it.  We use this for
# character conversion, for instance.
#
# Dana Jacobsen (dana@acm.org)
# 6 January 1995
#

package bp_csv;

######

&bib'reg_format(
  'csv',    # name
  'csv',     # short name
  'bp_csv', # package name
  'none',    # default character set
# our functions
  'open',
  'close is standard',
  'clear is standard',
  'options',
  'read',
  'write',
  'explode',
  'implode',
  'tocanon',
  'fromcanon',
);

######

#
# This controls how many lines to read in at a time.  Values less than 1
# indicate we are to read a paragraph at a time.
#
# Setting this to paragraph instead of line is a going to be a lot faster,
# as most of the time will be spent traversing subroutines otherwise.
#
$glb_inlines = -1;
$glb_firstline = 1 ;

######

sub options {
  local($opts) = @_;

  print "setting options to $opts\n";
}

######



sub open {
  local($file) = @_;

  &panic("open_stdbib called with no arguments")  unless defined $file;

  &bib'debugs("opening $file<$glb_current_fmt> ($glb_current_fh)", 128, 'module');

  if( open($bib'glb_current_fh, $file) ){
     $glb_firstline = 2 ;
#      print $bib'glb_current_fh "hello there\n" ;
      return $bib'glb_current_fmt  
  }      

  &goterror("Can't open file $file");
}



#####

sub read {
  local($file) = @_;

  return scalar(<$bib'glb_current_fh>) if $glb_inlines == 1;

  if ($glb_inlines < 1) {
    local($/) = '';
    return scalar(<$bib'glb_current_fh>);
  }

  local($ent) = '';
  local($loops) = $glb_inlines;
  while (<$bib'glb_current_fh>) {
    $ent .= $_;
    last unless --$loops;
  }
  $ent;
}

######

@can_list = 
   ('CiteKey',      'Identifier',
    'Title',        'Title',
    'SuperTitle',   'booktitle',
    'AuthorAddress','affiliation',
    'School',       'school',
    'Organization', 'organization',
    'Journal',      'journal',
    'ReportType',   'type',
    'Series',       'series',
    'Volume',       'volume',
    'Edition',      'edition',
    'Chapter',      'chapter',
    'Pages',        'pages',
    'PagesWhole',   'pages',
    'HowPublished', 'howpublished',
    'Publisher',    'publisher',
    'PubAddress',   'address',
    'Month',        'month',
    'Year',         'year',
    'Price',        'price',
    'Copyright',    'copyright',
    'Keywords',     'keywords',
    'MRNumber',     'mrnumber',
    'Language',     'language',
    'Annotation',   'annote',
    'ISBN',         'isbn',
    'ISSN',         'issn',
    'Field',        'subject',
    'Abstract',     'abstract',
    'Note',         'note',
    'Contents',     'contents',
    'Key',          'key',
    'Source',       'url',
   );



sub write {
  local($file, $out) = @_;

  if( $glb_firstline == 2 ) {
  print $bib'glb_current_fh ('"Identifier";"BibliographyType";"Author";"Title";"Year";"ISBN";"Booktitle";"Chapter";"Edition";"Editor";"Howpublished";"Institution";"Journal";"Month";"Note";"Annote";"Number";"Organizations";"Pages";"Publisher";"Address";"School";"Series";"ReportType";"Volume";"URL";"Custom1";"Custom2";"Custom3";"Custom4";"Custom5"'."\n") ;
  $glb_firstline = 1 ;
  }

  print $bib'glb_current_fh ($out,"\n");
}

######

sub explode {
  local($_) = @_;
  local(%entry) = ();

  $entry{'TEXT'} = $_;
  %entry;
}

######

sub implode {
  local(%can) = @_;
  
  $can{'CiteType'} = 0 ;

  '"'.$can{'CiteKey'}.'"'.";".
  '"'.$can{'CiteType'}.'"'.";".
  '"'.$can{'Authors'}.'"'.";".
  '"'.$can{'Title'}.'"'.";".
  '"'.$can{'Year'}.'"'.";".
  '"'.$can{'ISBN'}.'"'.";".
  '"'.$can{'SuperTitle'}.'"'.";".
  '"'.$can{'Chapter'}.'"'.";".
  '"'.$can{'Edition'}.'"'.";".
  '"'.$can{'Editors'}.'"'.";".
  '"'.$can{'Howpublished'}.'"'.";".
  '"'.$can{'School'}.'"'.";".
  '"'.$can{'Journal'}.'"'.";".
  '"'.$can{'Month'}.'"'.";".
  '"'.$can{'Note'}.'"'.";".
  '"'.$can{'Annote'}.'"'.";".
  '"'.$can{'Number'}.'"'.";".
  '"'.$can{'Organization'}.'"'.";".
  '"'.$can{'Pages'}.'"'.";".
  '"'.$can{'Publisher'}.'"'.";".
  '"'.$can{'Address'}.'"'.";".
  '"'.$can{'School'}.'"'.";".
  '"'.$can{'Series'}.'"'.";".
  '"'.$can{'ReportType'}.'"'.";".
  '"'.$can{'Volume'}.'"'.";".
  '"'.$can{'Source'}.'"'.";"
  ;
  
  
  
}

######

sub tocanon {
  local(%entry) = @_;
  %entry;
}

######

sub fromcanon {
  local(%entry) = @_;
  %entry;
}

#######################
# end of package
#######################

1;
