#
# bibliography package for Perl
#
# University of California Library -- Melvyl
#
# Dana Jacobsen (dana@acm.org)
# 8 November 1995
#
# Modified substantially by David Kaplan (dmkaplan@ucdavis.edu)
# 10 August, 2003.
#
# This is updated to the format of melvyl for 2003.  Melvyl seems to
# keep changing its format, so it is hard to keep up.  Fortunately, they
# seem to be just doing books at the moment, which makes it easier.
#

package bp_melvyl;

$version = "melvyl (dj 8 nov 95)";

######

&bib::reg_format(
  'melvyl',     # name
  'mel',        # short name
  'bp_melvyl',  # package name
  'none',       # default character set
  'suffix is mel',
# our functions
  'options is standard',
  'open is standard',
  'close is standard',
  'read',
  'write is standard',
  'clear is standard',
  'explode',
  'implode is unsupported',
  'tocanon',
  'fromcanon is unsupported',
);

######################################################################
# Some standard conversions.
###################################################################### 
%melvyl_fields = ( 
		   "AU" => [ "Authors", "$bib::cs_sep", 0 ],
		   "OA" => [ "Authors", "$bib::cs_sep", 0 ],
		   "LO" => [ "Location", "\n", 1 ],
		   "PB" => [ "Publisher", "", 1 ],
		   "PL" => [ "Address", ", ", 1 ],
		   "PH" => [ "Pages", "", 0 ],
		   "DP" => [ "Year", "", 1 ],
		   "SU" => [ "Keywords", "; ", 1 ],
		   "MT" => [ "Title", " ", 1 ],
		   "ED" => [ "Edition", "", 0 ],
		   "LG" => [ "Language", "; ", 1 ],
		   "NT" => [ "Notes", "\n" ],
		   "PT" => [ "Publication_Type", "", 1 ],
		   );

%melvyl_types = ( "Book" => "book" );


######################################################################
# read'
# This function is necessary as melvyl now seperates records with ZZ 
# Automatically drops certain lines.
#
######################################################################

sub read {
  local($file) = @_;

  local($ent) = '';
  while (<$bib::glb_current_fh>) {
      next if /^\s*$/; # skip empty lines.

      s/^\s{5}//; # Remove annoying extra spaces at start of line.
      last if /^ZZ/; # Go in if end.
      
      next unless /[A-Z]{2}|\s+/; # Go on if not a useful line.

      $ent .= $_;

  }

  $ent;
}


######

sub explode {
  my (%entry) = ();
  my ($field) = undef;
  my ($ent) = shift;

  foreach (split(/\n/,$ent)) {
      if (/^\s+/) {
	  return &bib::goterror("Melvyl explode--Problems parsing entry") unless defined $field;

	  s/^\s+//; # Strip spaces at beginning of line.
	  s/\s*$//; # Get rid of trailing periods and commas and spaces.

	  # Add line to entry.  Melvyl fields always come with first line.
	  if (defined($melvyl_fields{$field})) {
	      $entry{$field} .= $melvyl_fields{$field}[1] . $_;
	  } else {
	      $entry{$field} .= " " . $_;
	  }

	  next;
      } else { # Deal with new field type.
	  /^([A-Z]{2})\s+(.+)\s*$/;
	  
	  # Save all fields.
	  $field = $1; # If it is known.
	  $entry{$field} = $2;

      }
  }

  # Remove trailing junk if desired.
  foreach (keys(%entry)) {
      if (defined($melvyl_fields{$_}) && $melvyl_fields{$_}[2]) {
	  $entry{$_} =~ s/\s*[\/,\.]$//;
      }
  }
  
  %entry;
}

######


sub implode {
  local(%entry) = @_;
  return &bib::goterror("melvyl implode isn't supported.");
}

######

sub tocanon {
  my(%entry) = @_;
  my(%can,%s,$k);

  # Deal with special cases.
  $can{"CiteType"} = parse_melvyl_type( $entry{"PT"} );

  # Deal with regular entries.
  foreach (keys(%entry)) {
      if (defined($melvyl_fields{$_})) {
	  # Known field.

	  if ( defined($can{$melvyl_fields{$_}[0]}) ) {
	      # Add on with appropriate separator.  Basically for names.
	      $can{$melvyl_fields{$_}[0]} .= $melvyl_fields{$_}[1] . $entry{$_};
	  } else {
	      $can{$melvyl_fields{$_}[0]} = $entry{$_};
	  }
      } else {
	  # Not known.
	  $can{$_} = $entry{$_};
      }
  }
      
  # Parse names.
  $can{"Authors"} = parse_melvyl_names( $can{"Authors"} );
  $can{"Editors"} = parse_melvyl_names( $can{"Editors"} ) 
      if defined($can{"Editors"});

  %can;

}

######################################################################
# Tries to parse melvyl names, but this is hard.
######################################################################

sub parse_melvyl_names {
    my ($n) = @_;
    my (@nn) = ();

    foreach (split(/$bib::cs_sep/,$n)) { 
	my (@fl) = split(/,\s*/);

	# Deal with Junior's.
	if (defined($fl[2]) && $fl[2] =~ /([sSjJ]r\.?|I{2,})/) {
	    $fl[2] = ", $1";
	} else {
	    $fl[2] = "";
	}

	push( @nn, bp_util::name_to_canon( "$fl[1] $fl[0]$fl[2]" ) );
    }

    $n = join( $bib::cs_sep, @nn ); # Join them together.
}

######################################################################
# Tries to figure out type of reference. NEED more examples.
######################################################################

sub parse_melvyl_type {
    my ($t) = shift;
    
    if ( defined( $melvyl_types{$t} ) ) {
	$t = $melvyl_types{$t};
    } else {
	$t = "misc";
    }

    $t;
}

#######################
# end of package
#######################

1;
