#
# bibliography package for Perl
#
# University of California Library -- Melvyl
#
# Dana Jacobsen (dana@acm.org)
# 8 November 1995
#
#        This has not been tested yet.
#

package bp_melvyl_old;

$version = "melvyl (dj 8 nov 95)";

######

&bib::reg_format(
  'melvyl_old',     # name
  'mel_old',        # short name
  'bp_melvyl_old',  # package name
  'none',       # default character set
  'suffix is mel',
# our functions
  'options is standard',
  'open is standard',
  'close is standard',
  'read',
  'write is standard',
  'clear is standard',
  'explode',
  'implode is unsupported',
  'tocanon',
  'fromcanon is unsupported',
);

######################################################################
# read'
# This function is necessary as melvyl now seperates records with ZZ 
# Automatically drops certain lines.
#
######################################################################

sub read {
  local($file) = @_;

  local($ent) = '';
  while (<$bib::glb_current_fh>) {
    $ent .= $_ unless ( /^[0-9]+\.\s+/ );
    last if /^ZZ/;
  }

  $ent;
}


######

sub explode {
  local($_) = @_;
  local(%entry) = ();
  local($val);

  local($field) = undef;
  local(@lines) = split(/\n/);
  foreach (@lines) {
    if (/^\s/) {
      return &bib::goterror("Melvyl explode--Problems parsing entry") unless defined $field;
      s/^\s+/$bib::cs_sep2/; # This will later be dealt with, depending on field.
      s/\s+$//;
      $entry{$field} .= $_;
      next;
    }
    next unless /^[A-Z][A-Z]\s/;
    last if /^ZZ\s*$/;
    ($field, $val) = /^([A-Z]+)\s+(.*)/;
    $val =~ s/^\s+//;
    $val =~ s/\s+$//;
    if (defined $entry{$field}) {
      $entry{$field} .= $bib::cs_sep . $val;
    } else {
      $entry{$field} = $val;
    }
  }


  %entry;
}

######


sub implode {
  local(%entry) = @_;
  return &bib::goterror("melvyl implode isn't supported.");
}

######

sub tocanon {
  local(%entry) = @_;
  local(%can);
  local($mon, $yr);

  # Deal with multiple lines correctly.
  $entry{'AA'} =~ s/$bib::cs_sep2/ /g;
  $entry{'AT'} =~ s/$bib::cs_sep2/ /g;
  $entry{'MT'} =~ s/$bib::cs_sep2/ /g;
  $entry{'JT'} =~ s/$bib::cs_sep2/ /g;
  $entry{'AB'} =~ s/$bib::cs_sep2/ /g;
  $entry{'PA'} =~ s/$bib::cs_sep2/ /g;
  $entry{'PU'} =~ s/$bib::cs_sep2/ /g;
  $entry{'AK'} =~ s/$bib::cs_sep2/ /g;

  # TX and EL are special - don't want spaces in URL.
  $entry{'TX'} =~ s/$bib::cs_sep2//g;
  $entry{'EL'} =~ s/$bib::cs_sep2//g;

  # DT
  if (defined $entry{'DT'}) {
    if ($entry{'DT'} =~ /^ARTICLE/) {
      $can{'CiteType'} = 'article';
    } elsif ($entry{'DT'} =~ /JOURNAL ARTICLE/) {
      # This is a hack - it appears that JOURNAL ARTICLE can appear after something else.
      # eg. CONFERENCE LITERATURE; JOURNAL ARTICLE
      $can{'CiteType'} = 'article';
    } elsif ($entry{'DT'} =~ /^CONFERENCE PAPER/) {
      $can{'CiteType'} = 'inproceedings';
    } elsif ($entry{'DT'} =~ /^BOOK/) {
      $can{'CiteType'} = 'book';
    } elsif ($entry{'DT'} =~ /^DISSERTATION/) {
      $can{'CiteType'} = 'thesis';
    } elsif ($entry{'DT'} =~ /^GOVERNMENT DOCUMENT/) {
      $can{'CiteType'} = 'report';
    } else {
      $can{'CiteType'} = 'misc';
    }
  } else {
     &bib::gotwarn("missing type tag");
    $can{'CiteType'} = 'misc';
  }

  # PA
  if (defined $entry{'PA'}) {
    $can{'Authors'} = '';
    foreach $n (split(/$bib::cs_sep/, $entry{'PA'})) {
      $can{'Authors'} .= $bib::cs_sep . &melname_to_canon($n);
    }
    $can{'Authors'} =~ s/^$bib::cs_sep//;
  }

  $can{'Keywords'} = $entry{'SU'}  if defined $entry{'SU'};

  ($mon, $yr) = &bp_util::parsedate($entry{'DP'});
  $can{'Month'}         = $mon if $mon !~ /^\s*$/;
  $can{'Year'}          = $yr  if defined $yr;

  $can{'AuthorAddress'} = $entry{'AA'}  if defined $entry{'AA'};
  $can{'PubAddress'}    = $entry{'PL'}  if defined $entry{'PL'};
  $can{'Title'}         = $entry{'AT'}  if defined $entry{'AT'};
  $can{'Journal'}       = $entry{'JT'}  if defined $entry{'JT'};
  $can{'Publisher'}     = $entry{'PU'}  if defined $entry{'PU'};
  $can{'Series'}        = $entry{'SE'}  if defined $entry{'SE'};
  $can{'Abstract'}      = $entry{'AB'}  if defined $entry{'AB'};

  # Do some stripping on volume and number so they look better.
  $entry{'VO'} =~ s/^\s*[Vv]ol\.?\s*//;
  $entry{'NR'} =~ s/^\s*[Nn](o|um)\.?\s*//;
  $entry{'VO'} =~ s/[\.\,\;\:]?\s*$//;
  $entry{'NR'} =~ s/[\.\,\;\:]?\s*$//;
  $can{'Volume'}        = $entry{'VO'}  if defined $entry{'VO'};
  $can{'Number'}        = $entry{'NR'}  if defined $entry{'NR'};

  # Things that will go in misc.
  $can{'Language'}      = $entry{'LA'}  if defined $entry{'LA'};
  $can{'Database'}      = "Melvyl:" . $entry{'DB'}  if defined $entry{'DB'};
  $can{'UniqueID'}            = $entry{'AN'}  if defined $entry{'AN'};

  # Replace cs_sep with semi-colon in keywords.
  $entry{'AK'} =~ s/$bib::cs_sep/\;/g;
  $can{'Keywords'}      = $entry{'AK'} if defined $entry{'AK'};

  # Extract URL from TX entry.
  $entry{'TX'} =~ /(http[^\s]*)/;
  $entry{'TX'} = $+ if (defined($+) && $+); # Reduce entry to URL if pattern matched and is not empty.
  $can{'url'} = $entry{'TX'} if defined $entry{'TX'};

  # Extract URL from EL entry.
  $entry{'EL'} =~ /(http[^\s]*)/;
  $entry{'EL'} = $+ if (defined($+) && $+); # Reduce entry to URL if pattern matched and is not empty.
  $can{'url'} = $entry{'EL'} if defined $entry{'EL'};

  if (defined $entry{'MT'}) {
    if (defined $entry{'AT'}) {
      $can{'SuperTitle'} = $entry{'MT'};
    } else {
      $can{'Title'} = $entry{'MT'};
    }
  }

  # Strip date of any trailing .'s.
  $entry{'PG'} =~ s/\.$//;

  if ($can{'CiteType'} =~ /article/) {
    $can{'Pages'}         = $entry{'PG'}  if defined $entry{'PG'};
  } else {
    $can{'PagesWhole'}    = $entry{'PG'}  if defined $entry{'PG'};
  }

  $can{'OrigFormat'} = $version;
  %can;
}

sub melname_to_canon {
  local($name) = @_;
  local($last, $first, $cname);

  if ( defined $entry{'DB'} && ($entry{'DB'} =~ /MEDLINE/) )
  {
    # medline
    ($last, $first) = $name =~ /^(.*)\s+([A-Z]+)$/;
    if (defined $first) {
      $first =~ s/([A-Z])/$1. /g;
      $first =~ s/\s+$//;
    } 
  } else {
    $name =~ /^\s*(.*),\s*(.*)\s*$/;
    ($last, $first) = ($1, $2);
  }

  # If all else fails, just use given name.
  # Needed for some names where last and first not clear.
  $last = $name if (!$last);

  $cname = $last . $bib::cs_sep2 . $bib::cs_sep2 . $first . $bib::cs_sep2;
  $cname;
}

######


#######################
# end of package
#######################

1;
