#
# bibliography package for Perl
#
# Ovid
#
# This conversion module if very much alpha and very fragile.  
#
# It seems that the ovid format varies greatly and is very hard
# to write a filter for.  Also, I think this filter has a lot of 
# repeated code.  This could be fixed if I had a better understanding
# of the Ovid format.
#
# When in doubt, consult the source - the OvidSource that is.
#
# Copyright David M. Kaplan 2002
# License: GPL
#

package bp_ovid;

$version = "ovid (dmk 13 sep 2002)";

######

&bib::reg_format(
  'ovid',     # name
  'ovid',        # short name
  'bp_ovid',  # package name
  'none',       # default character set
  'suffix is ovid',
# our functions
  'options is standard',
  'open is standard',
  'close is standard',
  'read',
  'write is standard',
  'clear is standard',
  'explode',
  'implode is unsupported',
  'tocanon',
  'fromcanon is unsupported',
);

######################################################################
# read'
# This function breaks up records on <#>'s.
#
######################################################################

sub read {
  local($file) = @_;

  local($ent) = '';
  while (<$bib::glb_current_fh>) {
      last if (/^\<[0-9]+\>/ and $ent);
      $ent .= $_;
  }

  $ent;
}


######

sub explode {
  local($_) = @_;
  local(%entry) = ();
  local($val);

  local($field) = undef;
  local(@lines) = split(/\n/);
  foreach (@lines) {
      if (/^\s/) {
	  return &bib::goterror("Ovid explode--Problems parsing entry") unless defined $field;
	  s/^\s+//; # Strip spaces at beginning of line.
	  $entry{$field} && ( $entry{$field} .= $bib::cs_sep . $_ ) || ( $entry{$field} = $_ );
	  next;
      } else { # Deal with various Ovid field variations.
	  if (/^Author/) {
	      $field = "Authors";
	  } elsif (/^Editorial/) { # Need this to fix editor ambiguity.
	      $field = $_;
	  } elsif (/^Editor/) {
	      $field = "Editors";
	  } elsif (/^Collection Editor/) {
	      $field = "Editors";
	  } elsif (/^Title of Periodical/) {
	      $entry{"Publication Type"} = "Periodical";
	      $field = "Title";
	  } elsif (/^Title/) {
	      $field = "Title";
	  } else {
	      $field = $_;
	  }
	      
      }
  }
  
  %entry;
}

######


sub implode {
  local(%entry) = @_;
  return &bib::goterror("ovid implode isn't supported.");
}

######

sub tocanon {
  local(%entry) = @_;
  local(%can);
  local($mon, $yr);
  local($e);

  # Deal with multiple lines correctly.
  $entry{'Concept Codes'} =~ s/$bib::cs_sep/;/g;
  $entry{'Subject Headings'} =~ s/$bib::cs_sep/;/g;
  $entry{'Editors'} =~ s/$bib::cs_sep/ /g;
  $entry{'Source'} =~ s/$bib::cs_sep/ /g;
  $entry{'Abstract'} =~ s/$bib::cs_sep/ /g;
  $entry{'Title'} =~ s/$bib::cs_sep/ /g;

  # Type - this is hard.  Also parses source along the way.
  if (defined $entry{'Publication Type'}) {
      $_ = $entry{'Publication Type'};
      if ( /^([Jj]ournal ([Aa]rticle|[Pp]aper)|[Aa]rticle|[Ll]iterature\s+[Rr]eview)/ ) { # Trying to identify papers.
	  $can{'CiteType'} = 'article';
	  
	  ($can{'Journal'},$can{'Volume'},$can{'Number'},$can{'Year'},$can{'Pages'},$can{'Publisher'}) =
	      &ovid_parse_source($entry{'Source'},'article') if (defined($entry{'Source'}));

      } elsif (/^[pP]eriodical/) { # This if for MLA Periodicals Index. FRAGILE.
	  $can{'CiteType'} = 'book';
	  $can{'Periodical'} = 'TRUE';

	  # Deal with editorial.
	  if (defined($entry{'Editorial Address'})) {
	      $can{'PubAddress'} = $entry{'Editorial Address'};

	      # This does a decent job of getting publisher - not perfect.
	      $entry{'Editorial Address'} =~ /^([^,]+)/;
	      $can{'Publisher'} = $1;
	  }
	  
      } elsif (/^[Cc]onference/ || /^[Mm]eeting/) {
	  $can{'CiteType'} = 'inproceedings';
	  
	  ($can{'SuperTitle'},$can{'Volume'},$can{'Number'},$can{'Year'},$can{'Pages'},$can{'Publisher'}) =
	      &ovid_parse_source($entry{'Source'},'inproceedings') if (defined($entry{'Source'}));

      } elsif (/^[Bb]ook [Aa]rticle/) {
	  $can{'CiteType'} = 'inbook';
	  
	  $entry{'Book Citation'} =~ /(^[^\.]+)/;
	  $can{'SuperTitle'} = $1;

	  ($can{'Publisher'},$can{'PubAddress'},$can{'Pages'},$can{'Year'}) =
	      &ovid_parse_source($entry{'Source'},'inbook') if (defined($entry{'Source'}));
      } elsif (/^[Bb]ook/) {
	  $can{'CiteType'} = 'book';

	  ($can{'Publisher'},$can{'PubAddress'},$can{'Pages'},$can{'Year'}) =
	      &ovid_parse_source($entry{'Source'},'book') if (defined($entry{'Source'}));
	  
      } else {
	  &bib::gotwarn("Unknown type tag: \"$entry{'Publication Type'}\".  Defaulting to misc.");
	  $can{'CiteType'} = 'misc';
      }
  } else {
      &bib::gotwarn("missing type tag - defaulting to article.");
      $can{'CiteType'} = 'article';

      ($can{'Journal'},$can{'Volume'},$can{'Number'},$can{'Year'},$can{'Pages'},$can{'Publisher'}) =
	  &ovid_parse_source($entry{'Source'},'article') if (defined($entry{'Source'}));

  }

  # Author names.
  if (defined $entry{'Authors'}) {
      $can{'Authors'} = &ovidname_to_canon($entry{'Authors'});
  }

  # Editors.
  if (defined $entry{'Editors'}) {
      $can{'Editors'} = &ovidname_to_canon($entry{'Editors'});
  }

  # Work up key words.
  $can{'Keywords'} = '';
  $can{'Keywords'} .= $entry{'Major Concepts'}  if defined $entry{'Major Concepts'};
  $can{'Keywords'} .= '; ' . $entry{'Miscellaneous Descriptors'}  if defined $entry{'Miscellaneous Descriptors'};
  $can{'Keywords'} .= '; ' . $entry{'Subject Headings'}  if defined $entry{'Subject Headings'};
  $can{'Keywords'} .= '; ' . $entry{'Key Phrase Identifiers'}  if defined $entry{'Key Phrase Identifiers'};

  # Abstract
  $can{'Abstract'} = $entry{'Abstract'} if (defined($entry{'Abstract'}));

  # Title - strip trailing period.
  $entry{'Title'} =~ s/\.$//;
  $can{'Title'} = $entry{'Title'} if (defined($entry{'Title'}));

  # Author address
  $can{'AuthorAddress'} = $entry{'Institution'}  if defined $entry{'Institution'};

  # Things that will go in misc.
  $can{'AccessionNumber'} = $entry{'Accession Number'}  if defined $entry{'Accession Number'};
  $can{'OvidSource'} = $entry{'Source'}  if defined $entry{'Source'};
  $can{'ClassificationCodes'} = $entry{'Classification Codes'}  if defined $entry{'Classification Codes'};
  $can{'ConferenceInformation'} = $entry{'Conference Information'}  if defined $entry{'Conference Information'};
  $can{'BookCitation'} = $entry{'Book Citation'}  if defined $entry{'Book Citation'};
  $can{'ConceptCodes'} = $entry{'Concept Codes'}  if defined $entry{'Concept Codes'};
  $can{'Language'} = $entry{'Language'}  if defined $entry{'Language'};
  $can{'year'} = $entry{'Year of Publication'}  if defined $entry{'Year of Publication'};

  $can{'OrigFormat'} = $version;

  # Strip any remaining empty space from the beginning and end of entries.
  foreach $e ( keys(%can) ) {
      $can{$e} =~ s/^\s+//;
      $can{$e} =~ s/\s+$//;
  }

  # Return
  %can;
}

# This function does the difficult task of parsing ovid author names.
# This function is very fragile and could be improved if I had a better
# idea of what the Ovid format was.
sub ovidname_to_canon {
  $_ = shift;
  local(@names, @f);
  local($n, $last, $first, $cname, $ntype);

  # First split up names.  Fragile.
  if (/\;\s+/) { 
      s/\[a\]//g; # Remove weird [a]
      @names = split(/\;\s+/);
      $ntype = "spaces";
  } elsif (/\.\s\s/) {
      @names = split(/.\s\s/);
      $ntype = "spaces";
  } elsif (/\,\-/) {
      @names = split(/\,\s/);
      $ntype = "dashes";
  } elsif (/%%%/) { # Multiline editors.
      s/\s*\(ed\.\)\s*//g; # Remove extraneous (.ed)
      @names = split(/%%%/);
      $ntype = "commas";
  } elsif (/\,/) { # Single author with name separated by ,
      $names[0] = $_;
      $ntype = "commas"; 
  } elsif (/ /) {
      $names[0] = $_;
      $ntype = "spaces";
  } else {
      $names[0] = $_;
      $ntype = "none";
  }

  $cname = '';

  # This part tries to parse names.  This is very fragile and could
  # probably be changed to a much simpler system.
  foreach $n ( @names ) {
      if ( $ntype eq "none" ) {
	  $last = $n;
	  $first = '';
      } elsif ( $ntype eq "commas" ) {
      	  $n =~ /^([^,]+),\s*(.*)/ || $n =~ /^(.*)()/;
	  ($last,$first) = ($1,$2);
      } elsif ( $ntype eq "spaces" ) {
	  $n =~ /^([^\s]+)\s+(.*)/ || $n =~ /^(.*)()/;
	  ($last,$first) = ($1,$2);
      } elsif ( $ntype =~ /dashes/ ) {
	  $n =~ s/-/ /g;
	  $n =~ /^([^,]+),\s*(.*)/ || $n =~ /^(.*)()/;
	  ($last,$first) = ($1,$2);
      }	  
      
      # Try to fix initials problem.
      @f = split(/\s+/,$first);
      if ( ($#f == 0) and ( $f[0] =~ /^[A-Z]{2}$/ ) ) {
	  @f = split(//,$f[0]); 
      }
      for $i ( 0 .. $#f ) {
	  $f[$i] .= '.' if ($f[$i] =~ /^[A-Z]$/);
      }
      $first = join(" ",@f);

      # Put it all together.
      $cname .= $last . $bib::cs_sep2 . $bib::cs_sep2 . $first . $bib::cs_sep2 . $bib::cs_sep;
	   
  }
  
  $cname;

}


# This function processes the source field
# Very hard and fragile.
sub ovid_parse_source {

    my $s = shift;
    my $t = shift;
    my @res;

    if ($t eq 'article') {
	if ($s =~ /\[print\]/) {
	    @res = split( /\.\s+/, $s );

	    # Pages.
	    $res[4] = $res[3];

	    # Year
	    $res[2] =~ /(18|19|20)([0-9]{2})/;
	    $res[3] = $1 . $2;

	    # Volume and Number.
	    $res[1] =~ s/\[print\]\s*//;
	    $res[1] =~ /(\w+)(\([\w\-\s]+\)){0,1}/;
	    ($res[1], $res[2]) = ( $1, $2 );
	    $res[2] =~ tr/\(\)//d;

	} elsif ($s =~ /(\w+)(\([\w\-\s]+\)){0,1}:([0-9]+\-[0-9]+)/) {
	    @res = ( '', $1, $2, '', $3, '' );

	    # Fix number
	    $res[2] =~ tr/\(\)//d;
	    
	    # Journal.
	    $s =~ /^([^.]+)/;
	    $res[0] = $1;

	    # Year. Works - sort of.
	    $s =~ /(18|19|20)([0-9]{2})/;
	    $res[3] = $1 . $2;

	} elsif ($s =~ /vol\..*no\./) {
	    @res = split(/\,\s+/,$s);
	    
	    # Volume.
	    $res[1] =~ s/vol\.//;

	    # Number
	    $res[2] =~ s/no\.//;

	    # Year.
	    $res[3] =~ /([0-9]{4})/;
	    $res[3] = $1;

	    # Pages and publisher.
	    $res[4] =~ /pp\.([0-9\-]+)\.\s+(Publisher:\s+\w+){0,1}/;
	    $res[4] = $1;
	    $res[5] = $2;
	    $res[5] = s/Publisher:\s+//;
	} else { # Some biosis refs are like this.	
	    @r = split(/\.\s+/,$s);

	    # Journal
	    @res = ( $r[0] ); 
	    
	    # Volume and number
	    $r[1] =~ /^([0-9]+)(\([0-9]+\)){0,1}/;
	    $res[1] = $1;
	    $res[2] = $2;
	    $res[2] =~ y/\(\)//d; # Get rid of parens
	    
	    # Year
	    $r[2] =~ /([\w.]*),*\s*(18|19|20)([0-9]{2})/; 
	    $res[3] = $2.$3;
	    print STDERR "month: $1\n\n";

	    # Pages
	    $res[4] = $r[3];

	    # No publisher
	    $res[5] = undef;
	}

    } elsif ($t eq 'inproceedings') {
	if ($s =~ /vol\..*no\./) {
	    @res = split(/\,\s+/,$s);
	    
	    # Volume.
	    $res[1] =~ s/vol\.//;
	    
	    # Number
	    $res[2] =~ s/no\.//;
	    
	    # Year.
	    $res[3] =~ /([0-9]{4})/;
	    $res[3] = $1;
	    
	    # Pages and publisher.
	    $res[4] =~ /pp\.([0-9\-]+)\.\s+(Publisher:\s+\w+){0,1}/;
	    $res[4] = $1;
	    $res[5] = $2;
	    $res[5] = s/Publisher:\s+//;
	} else {	
	    @r = split(/\.\s+/,$s);
	
	    # Year and pages.
	    $r[2] =~ /([0-9]{4}),\s+pp\.([0-9\-]+)/;
	    $r[2] = $1;
	    $r[3] = $2;
	    
	    @res = ( $r[0], undef, undef, $r[2], $r[3], undef ); 
	}

    } elsif ($t eq 'inbook') {
	@res = split(/[.;]\s+/,$s);
	
	my @r = split(/,\s+/,$res[2]);
	$res[2] = $r[0];
	$res[3] = $r[1];
    } else { # book
	my @r = split(/\.\s+/,$s);
	$res[0] = $r[0];
	$res[1] = $r[2];
	
	$r[1] =~ /^\s*([0-9]+),\s*pp\.\s*(.*)\s*$/;
	$res[2] = $2;
	$res[3] = $1;

    }

    @res;
    
}
	   

######


#######################
# end of package
#######################

1;
